# Modules dependencies
import pandas as pd
import numpy as np
import os


def coordinate_flattening(
        grid_data,
        grid_model_height=5,
        xc_coordinates_label="XC",
        yc_coordinates_label="YC",
        zc_coordinates_label="elevation",
        print_info=False):

    # Group per x-y coordinates
    xy_data_grouped = grid_data.groupby([xc_coordinates_label, yc_coordinates_label])

    # Some data to retrieve from xy grouped coordinates
    ## Max z coordinates
    xy_data_grouped_elevation_max = xy_data_grouped[zc_coordinates_label].max()
    xy_data_grouped_elevation_min = xy_data_grouped[zc_coordinates_label].min()
    ## z coordinate ranges on each xy group
    xy_elevation_ranges = xy_data_grouped_elevation_max - xy_data_grouped_elevation_min
    ## Maximum range and x-y location
    x_coord_max, y_coord_max = xy_elevation_ranges.idxmax()
    max_column_height = xy_elevation_ranges.max()
    ## Cuboids data at maximal range xy position
    max_column_data = grid_data.loc[(
        grid_data[xc_coordinates_label] == x_coord_max) \
        & (grid_data[yc_coordinates_label] == y_coord_max)]
    ## Data along the maximum range column
    max_column_max_height = max_column_data[zc_coordinates_label].max()
    max_column_min_height =  max_column_data[zc_coordinates_label].min()
    max_column_cuboids = int(max_column_height/grid_model_height) + 1

    if print_info:
        # A brief resume and human readable information print
        if max_column_height % grid_model_height != 0:
            error_message = "Unexpected behavior"
            print(error_message)
        else:
            info_message_total_height = "Largest column has {} meters height between centers"
            info_message_total_cuboids = "Largest column has {} cuboids"
            info_message_largest_column_range = "Largest ranges from {} to {} in z axis"
            info_max_column_xy_coordinates = "Largest column is located at x:{} y:{}"
            print(info_message_total_height.format(max_column_height))
            print(info_message_total_cuboids.format(max_column_cuboids))
            print(info_max_column_xy_coordinates.format(x_coord_max, y_coord_max))
            print(info_message_largest_column_range.format(max_column_min_height, max_column_max_height))

    # Vertical flattening operations
    ## Vertical slots of coordinate flattening operation
    zc_slots_new_coordinates = np.linspace(
        start=max_column_min_height,
        stop=max_column_max_height,
        num=max_column_cuboids)
    ## Posible vertical coordinates, the key is the number of cuboids
    ## the value is the transformed z coordinates
    possible_coordinates = {}
    for n_cuboids in range(max_column_cuboids+1):
        positions = ((np.arange(n_cuboids)+1)/(n_cuboids+1)*max_column_cuboids).astype(int)
        possible_coordinates[n_cuboids] = zc_slots_new_coordinates[positions]
    ## Assamble new coordinate columns of z coordinates
    new_coordinates_column = pd.Series(index=grid_data.index, dtype='float64')
    for indices, data in xy_data_grouped:
        new_coordinates = possible_coordinates[len(data)]
        sorted_data = data[zc_coordinates_label].sort_values()
        new_coordinates_column.loc[sorted_data.index] = new_coordinates
    ## Return displacements
    displacements = new_coordinates_column - grid_data.loc[:, zc_coordinates_label]
    return displacements


def wall_coordinates_displacements(
        grid_data,
        grid_model_height=5,
        xc_coordinates_label="XC",
        yc_coordinates_label="YC",
        zc_coordinates_label="elevation",
        type='hanging wall',
        print_info=False):
    # Verify type required

    if not type in ['hanging wall', 'foot wall']:
        raise NotImplementedError("Coordinate change {} isn't implemented".format(type))

    grid_data = grid_data.loc[:,[xc_coordinates_label, yc_coordinates_label, zc_coordinates_label]]

    # Group per x-y coordinates
    xy_data_grouped = grid_data.groupby([xc_coordinates_label, yc_coordinates_label])

    # Some data to retrieve from xy grouped coordinates
    ## Max z coordinates
    xy_data_grouped_elevation_max = xy_data_grouped[zc_coordinates_label].max()
    xy_data_grouped_elevation_min = xy_data_grouped[zc_coordinates_label].min()
    ## z coordinate ranges on each xy group
    xy_elevation_ranges = xy_data_grouped_elevation_max - xy_data_grouped_elevation_min
    ## Maximum range and x-y location
    x_coord_max, y_coord_max = xy_elevation_ranges.idxmax()
    max_column_height = xy_elevation_ranges.max()
    ## Cuboids data at maximal range xy position
    max_column_data = grid_data.loc[(
        grid_data[xc_coordinates_label] == x_coord_max)
        & (grid_data[yc_coordinates_label] == y_coord_max)]
    ## Data along the maximum range column
    max_column_max_height = max_column_data[zc_coordinates_label].max()
    max_column_min_height = max_column_data[zc_coordinates_label].min()
    max_column_cuboids = int(max_column_height / grid_model_height) + 1

    if print_info:
        # A brief resume and human readable information print
        if max_column_height % grid_model_height != 0:
            error_message = "Unexpected behavior"
            print(error_message)
        else:
            info_message_total_height = "Largest column has {} meters height between centers"
            info_message_total_cuboids = "Largest column has {} cuboids"
            info_message_largest_column_range = "Largest ranges from {} to {} in z axis"
            info_max_column_xy_coordinates = "Largest column is located at x:{} y:{}"
            print(info_message_total_height.format(max_column_height))
            print(info_message_total_cuboids.format(max_column_cuboids))
            print(info_max_column_xy_coordinates.format(x_coord_max, y_coord_max))
            print(info_message_largest_column_range.format(max_column_min_height, max_column_max_height))

        ## Compute displacements
        if type == 'hanging wall':
            displacements = max_column_max_height - xy_data_grouped_elevation_max
        elif type == 'foot wall':
            displacements = max_column_min_height - xy_data_grouped_elevation_min

        displacements = displacements.reset_index()
        displacements.columns = [xc_coordinates_label, yc_coordinates_label, 'displacements']

        displacements = grid_data.merge(displacements, on=[xc_coordinates_label, yc_coordinates_label], how='left')
        return displacements.loc[:, 'displacements']
